from canvas_sdk.commands.base import _BaseCommand


class HistoryOfPresentIllnessCommand(_BaseCommand):
    """A class for managing a HPI command within a specific note."""

    class Meta:
        key = "hpi"
        commit_required_fields = ("narrative",)

    narrative: str | None = None

    @property
    def values(self) -> dict:
        """The HPI command's field values."""
        return {"narrative": self.narrative}



from canvas_sdk.commands.base import _BaseCommand


class PlanCommand(_BaseCommand):
    """A class for managing a Plan command within a specific note."""

    class Meta:
        key = "plan"
        commit_required_fields = ("narrative",)

    narrative: str = ""

    @property
    def values(self) -> dict:
        """The Plan command's field values."""
        return {"narrative": self.narrative}



from datetime import date
from enum import Enum, IntEnum

from typing_extensions import TypedDict

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class AllergenType(IntEnum):
    """An enum representing the type of allergen."""

    ALLERGEN_GROUP = 1
    MEDICATION = 2
    INGREDIENT = 6


class Allergen(TypedDict):
    """A TypedDict representing an allergen."""

    concept_id: int
    concept_type: AllergenType


class AllergyCommand(BaseCommand):
    """A class for managing an Allergy command within a specific note."""

    class Meta:
        key = "allergy"
        commit_required_fields = ("allergy",)

    class Severity(Enum):
        MILD = "mild"
        MODERATE = "moderate"
        SEVERE = "severe"

    allergy: Allergen | None = None
    severity: Severity | None = None
    narrative: str | None = None
    approximate_date: date | None = None

    @property
    def values(self) -> dict:
        """The Allergy command's field values."""
        return {
            "allergy": self.allergy,
            "severity": self.severity.value if self.severity else None,
            "narrative": self.narrative,
            "approximate_date": (
                self.approximate_date.isoformat() if self.approximate_date else None
            ),
        }



import json
import re
from enum import EnumType
from types import NoneType, UnionType
from typing import Union, get_args, get_origin

from canvas_sdk.base import Model
from canvas_sdk.commands.constants import Coding
from canvas_sdk.effects import Effect
from canvas_sdk.effects.protocol_card import Recommendation


class _BaseCommand(Model):
    class Meta:
        key = ""
        originate_required_fields = ("note_uuid",)
        edit_required_fields = ("command_uuid",)
        delete_required_fields = ("command_uuid",)
        commit_required_fields = ("command_uuid",)
        enter_in_error_required_fields = ("command_uuid",)

    def constantized_key(self) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", self.Meta.key).upper()

    note_uuid: str | None = None
    command_uuid: str | None = None

    def _get_effect_method_required_fields(self, method: str) -> tuple:
        base_required_fields: tuple = getattr(_BaseCommand.Meta, f"{method}_required_fields", ())
        command_required_fields = super()._get_effect_method_required_fields(method)
        return tuple(set(base_required_fields) | set(command_required_fields))

    @property
    def values(self) -> dict:
        return {}

    @property
    def coding_filter(self) -> Coding | None:
        """The coding filter used for command insertion in protocol cards."""
        return None

    @classmethod
    def _get_property_choices(cls, name: str, schema: dict) -> list[dict] | None:
        definition = schema.get("properties", {}).get(name, {})
        if not (choice_ref := next((a.get("$ref") for a in definition.get("anyOf", [])), None)):
            return None
        choice_key = choice_ref.split("#/$defs/")[-1]
        return schema.get("$defs", {}).get(choice_key, {}).get("enum")

    @classmethod
    def _get_property_type(cls, name: str) -> type | None:
        annotation = cls.model_fields[name].annotation
        origin = get_origin(annotation)

        # Handle Union types
        if origin is UnionType or origin is Union:
            annotation_args = get_args(annotation)
            # Filter out NoneType and take the first valid type
            annotation = next(
                (arg for arg in annotation_args if arg is not NoneType), annotation_args[0]
            )

        if type(annotation) is EnumType:
            return str

        return annotation

    @classmethod
    def command_schema(cls) -> dict:
        """The schema of the command."""
        base_properties = {"note_uuid", "command_uuid"}
        schema = cls.model_json_schema()
        required_fields: tuple = getattr(cls.Meta, "commit_required_fields", ())
        return {
            definition.get("commands_api_name", name): {
                "required": name in required_fields,
                "type": cls._get_property_type(name),
                "choices": cls._get_property_choices(name, schema),
            }
            for name, definition in schema["properties"].items()
            if name not in base_properties
        }

    def originate(self, line_number: int = -1) -> Effect:
        """Originate a new command in the note body."""
        self._validate_before_effect("originate")
        return Effect(
            type=f"ORIGINATE_{self.constantized_key()}_COMMAND",
            payload=json.dumps(
                {
                    "note": self.note_uuid,
                    "data": self.values,
                    "line_number": line_number,
                }
            ),
        )

    def edit(self) -> Effect:
        """Edit the command."""
        self._validate_before_effect("edit")
        return Effect(
            type=f"EDIT_{self.constantized_key()}_COMMAND",
            payload=json.dumps(
                {
                    "command": self.command_uuid,
                    "data": self.values,
                }
            ),
        )

    def delete(self) -> Effect:
        """Delete the command."""
        self._validate_before_effect("delete")
        return Effect(
            type=f"DELETE_{self.constantized_key()}_COMMAND",
            payload=json.dumps({"command": self.command_uuid}),
        )

    def commit(self) -> Effect:
        """Commit the command."""
        self._validate_before_effect("commit")
        return Effect(
            type=f"COMMIT_{self.constantized_key()}_COMMAND",
            payload=json.dumps({"command": self.command_uuid}),
        )

    def enter_in_error(self) -> Effect:
        """Mark the command as entered-in-error."""
        self._validate_before_effect("enter_in_error")
        return Effect(
            type=f"ENTER_IN_ERROR_{self.constantized_key()}_COMMAND",
            payload=json.dumps({"command": self.command_uuid}),
        )

    def recommend(self, title: str = "", button: str | None = None) -> Recommendation:
        """Returns a command recommendation to be inserted via Protocol Card."""
        if button is None:
            button = self.constantized_key().lower().replace("_", " ")
        command = self.Meta.key.lower()
        return Recommendation(title=title, button=button, command=command, context=self.values)



from django.db import models


class CareTeamMembershipStatus(models.TextChoices):
    """CareTeamMembershipStatus."""

    PROPOSED = "proposed", "Proposed"
    ACTIVE = "active", "Active"
    SUSPENDED = "suspended", "Suspended"
    INACTIVE = "inactive", "Inactive"
    ENTERED_IN_ERROR = "entered-in-error", "Entered in Error"


class CareTeamRole(models.Model):
    """CareTeamRole."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_careteamrole_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    active = models.BooleanField()

    def __str__(self) -> str:
        return self.display


class CareTeamMembership(models.Model):
    """CareTeamMembership."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_careteammembership_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="care_team_memberships", null=True
    )
    staff = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="care_team_memberships", null=True
    )
    role = models.ForeignKey(
        "v1.CareTeamRole", related_name="care_teams", on_delete=models.DO_NOTHING, null=True
    )
    status = models.CharField(choices=CareTeamMembershipStatus.choices)
    lead = models.BooleanField()
    role_code = models.CharField()
    role_system = models.CharField()
    role_display = models.CharField()

    def __str__(self) -> str:
        return f"id={self.id}"



import json
from typing import Any
from unittest.mock import Mock

import pytest
from requests import Response

from canvas_cli.utils.print import print


@pytest.mark.parametrize(
    "message", ["a simple message", ["an array", "of messages"], {"one": "test"}]
)
def test_print_json_outputs_valid_json(message: Any, capfd: pytest.CaptureFixture[str]) -> None:
    """Test the output of print is always valid json."""
    print.json(message)
    output, _ = capfd.readouterr()
    try:
        json.loads(output)
    except ValueError as exc:
        assert AssertionError(f"{output} is not valid json: {exc}")


def test_print_json_outputs_kwargs(capfd: pytest.CaptureFixture[str]) -> None:
    """Test the output of print contains all given kwargs."""
    print.json("A message", status_code=200, a_string="a_value", an_array=[1, 2], a_dict={"one": 2})
    output, _ = capfd.readouterr()
    try:
        json_dict = json.loads(output)

        assert json_dict.get("a_string") == "a_value"
        assert json_dict.get("an_array") == [1, 2]
        assert json_dict.get("a_dict") == {"one": 2}

    except ValueError as exc:
        assert AssertionError(f"{output} is not valid json: {exc}")


def test_print_overrides_default(capfd: pytest.CaptureFixture[str]) -> None:
    """Test using `print` defaults to Rich."""
    message = "Testing print"
    print(message)
    output, _ = capfd.readouterr()
    assert message + "\n" == output


def test_print_response_non_json_text(capfd: pytest.CaptureFixture[str]) -> None:
    """Test print.response with a non-json response."""
    response = Mock(spec=Response)
    response.status_code = 200
    response.text = "testing text"
    response.json.side_effect = json.JSONDecodeError("", "", 0)
    print.response(response)
    output, _ = capfd.readouterr()
    assert json.loads(output) == json.loads(
        '{"status_code": 200, "success": true, "message": "testing text"}'
    )


def test_print_response_json_text(capfd: pytest.CaptureFixture[str]) -> None:
    """Test print.response with a json response."""
    response = Mock(spec=Response)
    response.status_code = 201
    response.json.return_value = {"something": True}
    print.response(response)
    output, _ = capfd.readouterr()
    assert json.loads(output) == json.loads(
        '{"status_code": 201, "success": true, "message": {"something": true} }'
    )



import json
from typing import Any

from canvas_sdk.base import Model
from canvas_sdk.effects import Effect, EffectType


class _BaseEffect(Model):
    """
    A Canvas Effect that changes user behavior or autonomously performs activities on behalf of users.
    """

    patient_filter: dict | None = None

    class Meta:
        effect_type = EffectType.UNKNOWN_EFFECT

    @property
    def values(self) -> dict[str, Any]:
        return {}

    @property
    def effect_payload(self) -> dict[str, Any]:
        return {"data": self.values}

    def apply(self) -> Effect:
        self._validate_before_effect("apply")
        return Effect(type=self.Meta.effect_type, payload=json.dumps(self.effect_payload))



from canvas_sdk.commands.commands.prescribe import PrescribeCommand


class RefillCommand(PrescribeCommand):
    """A class for managing a Refill command within a specific note."""

    class Meta:
        key = "refill"
        commit_required_fields = (
            "fdb_code",
            "sig",
            "quantity_to_dispense",
            "type_to_dispense",
            "refills",
            "substitutions",
            "prescriber_id",
        )



from pathlib import Path

import pytest

from canvas_sdk.effects import Effect
from canvas_sdk.events import Event, EventRequest, EventType
from plugin_runner.plugin_runner import LOADED_PLUGINS


@pytest.mark.parametrize("install_test_plugin", ["test_render_template"], indirect=True)
def test_render_to_string_valid_template(
    install_test_plugin: Path, load_test_plugins: None
) -> None:
    """Test that the render_to_string function loads and renders a valid template."""
    plugin = LOADED_PLUGINS[
        "test_render_template:test_render_template.protocols.my_protocol:ValidTemplate"
    ]
    result: list[Effect] = plugin["class"](Event(EventRequest(type=EventType.UNKNOWN))).compute()
    assert "html" in result[0].payload


@pytest.mark.parametrize("install_test_plugin", ["test_render_template"], indirect=True)
def test_render_to_string_invalid_template(
    install_test_plugin: Path, load_test_plugins: None
) -> None:
    """Test that the render_to_string function raises an error for invalid templates."""
    plugin = LOADED_PLUGINS[
        "test_render_template:test_render_template.protocols.my_protocol:InvalidTemplate"
    ]
    with pytest.raises(FileNotFoundError):
        plugin["class"](Event(EventRequest(type=EventType.UNKNOWN))).compute()


@pytest.mark.parametrize("install_test_plugin", ["test_render_template"], indirect=True)
def test_render_to_string_forbidden_template(
    install_test_plugin: Path, load_test_plugins: None
) -> None:
    """Test that the render_to_string function raises an error for a template outside plugin package."""
    plugin = LOADED_PLUGINS[
        "test_render_template:test_render_template.protocols.my_protocol:ForbiddenTemplate"
    ]
    with pytest.raises(PermissionError):
        plugin["class"](Event(EventRequest(type=EventType.UNKNOWN))).compute()



import pytest
from pydantic import ValidationError
from typer.testing import CliRunner

from canvas_sdk.commands import (
    AssessCommand,
    DiagnoseCommand,
    GoalCommand,
    HistoryOfPresentIllnessCommand,
    MedicationStatementCommand,
    PlanCommand,
    PrescribeCommand,
    QuestionnaireCommand,
    ReasonForVisitCommand,
    StopMedicationCommand,
    UpdateGoalCommand,
)
from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.tests.test_utils import (
    fake,
    get_field_type,
    raises_none_error_for_effect_method,
    raises_wrong_type_error,
)

runner = CliRunner()


@pytest.mark.parametrize(
    "Command,fields_to_test",
    [
        (AssessCommand, ("condition_id", "background", "status", "narrative")),
        (
            DiagnoseCommand,
            ("icd10_code", "background", "approximate_date_of_onset", "today_assessment"),
        ),
        (
            GoalCommand,
            (
                "goal_statement",
                "start_date",
                "due_date",
                "achievement_status",
                "priority",
                "progress",
            ),
        ),
        (HistoryOfPresentIllnessCommand, ("narrative",)),
        (MedicationStatementCommand, ("fdb_code", "sig")),
        (PlanCommand, ("narrative", "command_uuid")),
        (
            PrescribeCommand,
            (
                "fdb_code",
                "icd10_codes",
                "sig",
                "days_supply",
                "type_to_dispense",
                "refills",
                "substitutions",
                "pharmacy",
                "prescriber_id",
                "note_to_pharmacist",
            ),
        ),
        (QuestionnaireCommand, ("questionnaire_id", "result")),
        (ReasonForVisitCommand, ("coding", "comment")),
        (StopMedicationCommand, ("medication_id", "rationale")),
        (
            UpdateGoalCommand,
            (
                "goal_id",
                "due_date",
                "achievement_status",
                "priority",
                "progress",
            ),
        ),
    ],
)
def test_command_raises_generic_error_when_kwarg_given_incorrect_type(
    Command: type[_BaseCommand],
    fields_to_test: tuple[str],
) -> None:
    """Test that Command raises a generic error when a kwarg is given an incorrect type."""
    for field in fields_to_test:
        raises_wrong_type_error(Command, field)

    for method in ["originate", "edit", "delete", "commit", "enter_in_error"]:
        raises_none_error_for_effect_method(Command, method)


@pytest.mark.parametrize(
    "Command,err_kwargs,err_msg,valid_kwargs",
    [
        (
            PlanCommand,
            {"narrative": "yo", "note_uuid": 1},
            "1 validation error for PlanCommand\nnote_uuid\n  Input should be a valid string [type=string_type",
            {"narrative": "yo", "note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            PlanCommand,
            {"narrative": "yo", "note_uuid": "5", "command_uuid": 5},
            "1 validation error for PlanCommand\ncommand_uuid\n  Input should be a valid string [type=string_type",
            {"narrative": "yo", "note_uuid": "5", "command_uuid": "5"},
        ),
        (
            ReasonForVisitCommand,
            {"note_uuid": "00000000-0000-0000-0000-000000000000", "structured": True},
            "1 validation error for ReasonForVisitCommand\n  Structured RFV should have a coding",
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "structured": False,
            },
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": "x"},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.system\n  Field required [type=missing",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": 1, "system": "y"},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.code\n  Input should be a valid string [type=string_type",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": None, "system": "y"},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.code\n  Input should be a valid string [type=string_type",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"system": "y"},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.code\n  Field required [type=missing",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": "x", "system": 1},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.system\n  Input should be a valid string [type=string_type",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": "x", "system": None},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.system\n  Input should be a valid string [type=string_type",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (
            ReasonForVisitCommand,
            {
                "note_uuid": "00000000-0000-0000-0000-000000000000",
                "coding": {"code": "x", "system": "y", "display": 1},
            },
            "1 validation error for ReasonForVisitCommand\ncoding.display\n  Input should be a valid string [type=string_type",
            {"note_uuid": "00000000-0000-0000-0000-000000000000"},
        ),
    ],
)
def test_command_raises_specific_error_when_kwarg_given_incorrect_type(
    Command: type[PlanCommand] | type[ReasonForVisitCommand],
    err_kwargs: dict,
    err_msg: str,
    valid_kwargs: dict,
) -> None:
    """Test that Command raises a specific error when a kwarg is given an incorrect type."""
    with pytest.raises(ValidationError) as e1:
        cmd = Command(**err_kwargs)
        cmd.originate()
        cmd.edit()
    assert err_msg in repr(e1.value)

    cmd = Command(**valid_kwargs)
    if len(err_kwargs) < len(valid_kwargs):
        return
    key, value = list(err_kwargs.items())[-1]
    with pytest.raises(ValidationError) as e2:
        setattr(cmd, key, value)
        cmd.originate()
        cmd.edit()
    assert err_msg in repr(e2.value)


@pytest.mark.parametrize(
    "Command,fields_to_test",
    [
        (AssessCommand, ("condition_id", "background", "status", "narrative")),
        (
            DiagnoseCommand,
            ("icd10_code", "background", "approximate_date_of_onset", "today_assessment"),
        ),
        (
            GoalCommand,
            (
                "goal_statement",
                "start_date",
                "due_date",
                "achievement_status",
                "priority",
                "progress",
            ),
        ),
        (HistoryOfPresentIllnessCommand, ("narrative",)),
        (MedicationStatementCommand, ("fdb_code", "sig")),
        (PlanCommand, ("narrative", "command_uuid", "note_uuid")),
        (
            PrescribeCommand,
            (
                "fdb_code",
                "icd10_codes",
                "sig",
                "days_supply",
                "quantity_to_dispense",
                "type_to_dispense",
                "refills",
                "substitutions",
                "pharmacy",
                "prescriber_id",
                "note_to_pharmacist",
            ),
        ),
        (QuestionnaireCommand, ("questionnaire_id", "result")),
        (ReasonForVisitCommand, ("coding", "comment")),
        (StopMedicationCommand, ("medication_id", "rationale")),
        (
            UpdateGoalCommand,
            (
                "goal_id",
                "due_date",
                "achievement_status",
                "priority",
                "progress",
            ),
        ),
    ],
)
def test_command_allows_kwarg_with_correct_type(
    Command: type[_BaseCommand],
    fields_to_test: tuple[str],
) -> None:
    """Test that Command allows a kwarg with the correct type."""
    schema = Command.model_json_schema()

    for field in fields_to_test:
        field_type = get_field_type(schema["properties"][field])

        init_field_value = fake({"type": field_type}, Command)
        init_kwargs = {field: init_field_value}
        cmd = Command(**init_kwargs)
        assert getattr(cmd, field) == init_field_value

        updated_field_value = fake({"type": field_type}, Command)
        setattr(cmd, field, updated_field_value)
        assert getattr(cmd, field) == updated_field_value

    for method in ["originate", "edit", "delete", "commit", "enter_in_error"]:
        required_fields = {
            k: v
            for k, v in schema["properties"].items()
            if k in Command()._get_effect_method_required_fields(method)
        }
        base = {field: fake(props, Command) for field, props in required_fields.items()}
        cmd = Command(**base)
        effect = getattr(cmd, method)()
        assert effect is not None



from django.db import models


class AppointmentProgressStatus(models.TextChoices):
    """AppointmentProgressStatus."""

    UNCONFIRMED = "unconfirmed", "Unconfirmed"
    ATTEMPTED = "attempted", "Attempted"
    CONFIRMED = "confirmed", "Confirmed"
    ARRIVED = "arrived", "Arrived"
    ROOMED = "roomed", "Roomed"
    EXITED = "exited", "Exited"
    NOSHOWED = "noshowed", "No-showed"
    CANCELLED = "cancelled", "Cancelled"


class Appointment(models.Model):
    """Appointment."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_appointment_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient",
        on_delete=models.DO_NOTHING,
        related_name="appointments",
        null=True,
    )
    appointment_rescheduled_from = models.ForeignKey(
        "self",
        on_delete=models.DO_NOTHING,
        related_name="appointment_rescheduled_to",
        null=True,
    )
    provider = models.ForeignKey("v1.Staff", on_delete=models.DO_NOTHING, null=True)
    start_time = models.DateTimeField()
    duration_minutes = models.IntegerField()
    comment = models.TextField(null=True)
    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING, null=True)

    note_type = models.ForeignKey(
        "v1.NoteType", on_delete=models.DO_NOTHING, related_name="appointments", null=True
    )

    status = models.CharField(
        max_length=20,
        choices=AppointmentProgressStatus,
    )
    meeting_link = models.URLField(null=True, blank=True)
    telehealth_instructions_sent = models.BooleanField()
    location = models.ForeignKey("v1.PracticeLocation", on_delete=models.DO_NOTHING, null=True)
    description = models.TextField(null=True, blank=True)



from ..value_set import ValueSet


class AceInhibitorOrArbIngredient(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for medications of angiotensin-converting enzyme (ACE) inhibitor and angiotensin-receptor blocker (ARB) ingredients.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance for prescribable angiotensin-converting enzyme (ACE) inhibitor and angiotensin-receptor blocker (ARB) ingredients only.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS135v10
    """

    VALUE_SET_NAME = "ACE Inhibitor or ARB Ingredient"
    OID = "2.16.840.1.113883.3.526.3.1489"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    RXNORM = {
        "1998",  # captopril
        "3827",  # enalapril
        "18867",  # benazepril
        "29046",  # lisinopril
        "30131",  # moexipril
        "35208",  # quinapril
        "35296",  # ramipril
        "38454",  # trandolapril
        "50166",  # fosinopril
        "52175",  # losartan
        "54552",  # perindopril
        "69749",  # valsartan
        "73494",  # telmisartan
        "83515",  # eprosartan
        "83818",  # irbesartan
        "214354",  # candesartan
        "321064",  # olmesartan
    }


class BetaBlockerTherapyIngredient(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for medications containing ingredients for beta blocker therapy.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance for prescribable beta blockers.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS144v10, CMS145v10
    """

    VALUE_SET_NAME = "Beta Blocker Therapy Ingredient"
    OID = "2.16.840.1.113883.3.526.3.1493"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    RXNORM = {
        "149",  # acebutolol
        "1202",  # atenolol
        "1520",  # betaxolol
        "2116",  # carteolol
        "6185",  # labetalol
        "6918",  # metoprolol
        "7226",  # nadolol
        "8332",  # pindolol
        "8787",  # propranolol
        "10600",  # timolol
        "19484",  # bisoprolol
        "20352",  # carvedilol
    }


class EggSubstance(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for an allergy or intolerance to an egg substance.

    **Data Element Scope:** This value set may use a model element related to Substance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance to an egg substance.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = "Egg Substance"
    OID = "2.16.840.1.113883.3.526.3.1537"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    SNOMEDCT = {
        "102263004",  # Eggs (edible) (substance)
        "226881001",  # Dried egg (substance)
        "226885005",  # Raw egg (substance)
        "229955000",  # Dried egg white (substance)
        "256442007",  # Egg yolk (substance)
        "256443002",  # Egg white (substance)
        "286550009",  # Hen's egg (substance)
        "303300008",  # Egg protein (substance)
        "414074006",  # Egg product (substance)
    }


class InfluenzaVaccination(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for an influenza vaccination.

    **Data Element Scope:** This value set may use a model element related to Procedure.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccination.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = "Influenza Vaccination"
    OID = "2.16.840.1.113883.3.526.3.402"
    DEFINITION_VERSION = "20170908"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CPT = {
        "90630",  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, for intradermal use
        "90653",  # Influenza vaccine, inactivated (IIV), subunit, adjuvanted, for intramuscular use
        "90654",  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative-free, for intradermal use
        "90655",  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative free, 0.25 mL dosage, for intramuscular use
        "90656",  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative free, 0.5 mL dosage, for intramuscular use
        "90657",  # Influenza virus vaccine, trivalent (IIV3), split virus, 0.25 mL dosage, for intramuscular use
        "90658",  # Influenza virus vaccine, trivalent (IIV3), split virus, 0.5 mL dosage, for intramuscular use
        "90661",  # Influenza virus vaccine, trivalent (ccIIV3), derived from cell cultures, subunit, preservative and antibiotic free, 0.5 mL dosage, for intramuscular use
        "90662",  # Influenza virus vaccine (IIV), split virus, preservative free, enhanced immunogenicity via increased antigen content, for intramuscular use
        "90666",  # Influenza virus vaccine (IIV), pandemic formulation, split virus, preservative free, for intramuscular use
        "90667",  # Influenza virus vaccine (IIV), pandemic formulation, split virus, adjuvanted, for intramuscular use
        "90668",  # Influenza virus vaccine (IIV), pandemic formulation, split virus, for intramuscular use
        "90673",  # Influenza virus vaccine, trivalent (RIV3), derived from recombinant DNA, hemagglutinin (HA) protein only, preservative and antibiotic free, for intramuscular use
        "90674",  # Influenza virus vaccine, quadrivalent (ccIIV4), derived from cell cultures, subunit, preservative and antibiotic free, 0.5 mL dosage, for intramuscular use
        "90682",  # Influenza virus vaccine, quadrivalent (RIV4), derived from recombinant DNA, hemagglutinin (HA) protein only, preservative and antibiotic free, for intramuscular use
        "90685",  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, 0.25 mL dosage, for intramuscular use
        "90686",  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, 0.5 mL dosage, for intramuscular use
        "90687",  # Influenza virus vaccine, quadrivalent (IIV4), split virus, 0.25 mL dosage, for intramuscular use
        "90688",  # Influenza virus vaccine, quadrivalent (IIV4), split virus, 0.5 mL dosage, for intramuscular use
        "90689",  # Influenza virus vaccine, quadrivalent (IIV4), inactivated, adjuvanted, preservative free, 0.25 mL dosage, for intramuscular use
        "90694",  # Influenza virus vaccine, quadrivalent (aIIV4), inactivated, adjuvanted, preservative free, 0.5 mL dosage, for intramuscular use
        "90756",  # Influenza virus vaccine, quadrivalent (ccIIV4), derived from cell cultures, subunit, antibiotic free, 0.5 mL dosage, for intramuscular use
    }
    HCPCSLEVELII = {
        "G0008",  # Administration of influenza virus vaccine
        "Q2034",  # Influenza virus vaccine, split virus, for intramuscular use (agriflu)
        "Q2035",  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (afluria)
        "Q2036",  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (flulaval)
        "Q2037",  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (fluvirin)
        "Q2038",  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (fluzone)
        "Q2039",  # Influenza virus vaccine, not otherwise specified
    }
    SNOMEDCT = {
        "86198006",  # Administration of vaccine product containing only Influenza virus antigen (procedure)
    }


class InfluenzaVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with influenza vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = "Influenza Vaccine"
    OID = "2.16.840.1.113883.3.526.3.1254"
    DEFINITION_VERSION = "20170908"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "88",  # influenza virus vaccine, unspecified formulation
        "135",  # influenza, high dose seasonal, preservative-free
        "140",  # Influenza, seasonal, injectable, preservative free
        "141",  # Influenza, seasonal, injectable
        "144",  # seasonal influenza, intradermal, preservative free
        "149",  # influenza, live, intranasal, quadrivalent
        "150",  # Influenza, injectable, quadrivalent, preservative free
        "155",  # Seasonal, trivalent, recombinant, injectable influenza vaccine, preservative free
        "158",  # influenza, injectable, quadrivalent, contains preservative
        "161",  # Influenza, injectable,quadrivalent, preservative free, pediatric
        "166",  # influenza, intradermal, quadrivalent, preservative free, injectable
        "168",  # Seasonal trivalent influenza vaccine, adjuvanted, preservative free
        "171",  # Influenza, injectable, Madin Darby Canine Kidney, preservative free, quadrivalent
        "185",  # Seasonal, quadrivalent, recombinant, injectable influenza vaccine, preservative free
        "186",  # Influenza, injectable, Madin Darby Canine Kidney,  quadrivalent with preservative
        "197",  # influenza, high-dose seasonal, quadrivalent, .7mL dose, preservative free
        "205",  # influenza, seasonal vaccine, quadrivalent, adjuvanted, .5mL dose, preservative free
    }


class StatinAllergen(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of an allergy to statin medications.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that define an allergy to statin medications.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS347v5
    """

    VALUE_SET_NAME = "Statin Allergen"
    OID = "2.16.840.1.113762.1.4.1110.42"
    DEFINITION_VERSION = "20210220"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    RXNORM = {
        "6472",  # lovastatin
        "36567",  # simvastatin
        "41127",  # fluvastatin
        "42463",  # pravastatin
        "83367",  # atorvastatin
        "301542",  # rosuvastatin
        "861634",  # pitavastatin
    }
    SNOMEDCT = {
        "372912004",  # Substance with 3-hydroxy-3-methylglutaryl-coenzyme A reductase inhibitor mechanism of action (substance)
    }



from enum import Enum
from typing import Any

from pydantic import Field

from canvas_sdk.effects.base import EffectType, _BaseEffect


class AddBannerAlert(_BaseEffect):
    """
    An Effect that will result in a banner alert in Canvas.
    """

    class Meta:
        effect_type = EffectType.ADD_BANNER_ALERT
        apply_required_fields = (
            "patient_id|patient_filter",
            "key",
            "narrative",
            "placement",
            "intent",
        )

    class Placement(Enum):
        CHART = "chart"
        TIMELINE = "timeline"
        APPOINTMENT_CARD = "appointment_card"
        SCHEDULING_CARD = "scheduling_card"
        PROFILE = "profile"

    class Intent(Enum):
        INFO = "info"
        WARNING = "warning"
        ALERT = "alert"

    patient_id: str | None = None
    key: str | None = None
    narrative: str | None = Field(max_length=90, default=None)
    placement: list[Placement] | None = Field(min_length=1, default=None)
    intent: Intent | None = None
    href: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The BannerAlert's values."""
        return {
            "narrative": self.narrative,
            "placement": [p.value for p in self.placement] if self.placement else None,
            "intent": self.intent.value if self.intent else None,
            "href": self.href,
        }

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {
            "patient": self.patient_id,
            "patient_filter": self.patient_filter,
            "key": self.key,
            "data": self.values,
        }



from django.db import models

from canvas_sdk.v1.data.common import (
    DocumentReviewMode,
    OrderStatus,
    ReviewPatientCommunicationMethod,
    ReviewStatus,
)


class ImagingOrder(models.Model):
    """Model to read ImagingOrder data."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_imagingorder_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="imaging_orders", null=True
    )
    # TODO - uncomment when Note model is complete
    #  note = models.ForeigneKey(Note, on_delete=models.DO_NOTHING, related_name="imaging_orders", null=True)
    imaging = models.CharField()
    # TODO - uncomment when ServiceProvider model is complete
    # imaging_center = models.ForeignKey('v1.ServiceProvider', on_delete=models.DO_NOTHING, related_name="imaging_orders", null=True)
    note_to_radiologist = models.CharField()
    internal_comment = models.CharField()
    status = models.CharField(choices=OrderStatus)
    date_time_ordered = models.DateTimeField()
    priority = models.CharField()
    # TODO - uncomment when Staff model is complete
    # ordering_provider = models.ForeignKey('v1.Staff', on_delete=models.DO_NOTHING, related_name="imaging_orders", null=True)
    delegated = models.BooleanField(default=False)


class ImagingReview(models.Model):
    """Model to read ImagingReview data."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_imagingreview_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient_communication_method = models.CharField(choices=ReviewPatientCommunicationMethod)
    # TODO  - uncomment when Note model is complete
    # note = models.ForeignKey('v1.Note', on_delete=models.DO_NOTHING, related_name="imaging_reviews", null=True)
    internal_comment = models.CharField()
    message_to_patient = models.CharField()
    is_released_to_patient = models.BooleanField()
    status = models.CharField(choices=ReviewStatus)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="imaging_reviews", null=True
    )


class ImagingReport(models.Model):
    """Model to read ImagingReport data."""

    class ImagingReportSource(models.TextChoices):
        RADIOLOGY_FROM_PATIENT = "RADIOLOGY_PATIENT", "Radiology Report From Patient"
        VERBAL_FROM_PATIENT = "VERBAL_PATIENT", "Verbal Report From Patient"
        DIRECTLY_REPORT = "DIRECTLY_RADIOLOGY", "Directly Radiology Report"

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_imagingreport_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    review_mode = models.CharField(choices=DocumentReviewMode)
    junked = models.BooleanField()
    requires_signature = models.BooleanField()
    assigned_date = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="imaging_results", null=True
    )
    order = models.ForeignKey(ImagingOrder, on_delete=models.DO_NOTHING, null=True)
    source = models.CharField(choices=ImagingReportSource)
    name = models.CharField()
    result_date = models.DateField()
    original_date = models.DateField()
    review = models.ForeignKey(ImagingReview, on_delete=models.DO_NOTHING, null=True)



from plugin_runner.plugin_runner import run_server


def run_plugin(plugin_directory: str) -> None:
    """
    Run the specified plugin for local development.
    """
    return run_plugins([plugin_directory])


def run_plugins(plugin_directories: list[str]) -> None:
    """
    Run the specified plugins for local development.
    """
    run_server(plugin_directories)
    return



import os

import arrow
from jwt import encode

from logger import log
from settings import PLUGIN_RUNNER_SIGNING_KEY

ONE_DAY_IN_MINUTES = 60 * 24


def token_for_plugin(
    plugin_name: str,
    audience: str,
    issuer: str = "plugin-runner",
    jwt_signing_key: str = PLUGIN_RUNNER_SIGNING_KEY,
    expiration_minutes: int = ONE_DAY_IN_MINUTES,
    extra_kwargs: dict | None = None,
) -> str:
    """
    Generate a JWT for the given plugin and audience.
    """
    if not extra_kwargs:
        extra_kwargs = {}

    if not jwt_signing_key:
        log.warning(
            "Using an insecure JWT signing key for GraphQL access. Set the PLUGIN_RUNNER_SIGNING_KEY environment variable to avoid this message."
        )

    token = encode(
        {
            "plugin_name": plugin_name,
            "customer_identifier": os.getenv("CUSTOMER_IDENTIFIER"),
            "exp": arrow.utcnow().shift(minutes=expiration_minutes).datetime,
            "aud": audience,
            "iss": issuer,
            **extra_kwargs,
        },
        jwt_signing_key,
        algorithm="HS256",
    )

    return token



from canvas_sdk.commands.commands.questionnaire import QuestionnaireCommand


class ReviewOfSystemsCommand(QuestionnaireCommand):
    """A class for managing physical exam command."""

    class Meta:
        key = "ros"
        commit_required_fields = ("questionnaire_id",)



from canvas_sdk.effects import Effect, EffectType
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol
from canvas_sdk.templates import render_to_string


class ValidTemplate(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    RESPONDS_TO = [EventType.Name(EventType.UNKNOWN)]

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        return [
            Effect(type=EffectType.LOG, payload=render_to_string("templates/template.html", None))
        ]


class InvalidTemplate(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    RESPONDS_TO = [EventType.Name(EventType.UNKNOWN)]

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        return [
            Effect(type=EffectType.LOG, payload=render_to_string("templates/template1.html", None))
        ]


class ForbiddenTemplate(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    RESPONDS_TO = [EventType.Name(EventType.UNKNOWN)]

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        return [
            Effect(
                type=EffectType.LOG,
                payload=render_to_string("../../templates/template.html", None),
            )
        ]



from django.apps import apps
from django.db import models


class Command(models.Model):
    """Command."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_commands_command_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    state = models.CharField()
    patient = models.ForeignKey("v1.Patient", on_delete=models.DO_NOTHING, null=True)
    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING, related_name="commands")
    schema_key = models.TextField()
    data = models.JSONField()
    origination_source = models.CharField()
    anchor_object_type = models.CharField()
    anchor_object_dbid = models.BigIntegerField()

    @property
    def anchor_object(self) -> models.Model | None:
        """
        Use the anchor_object_type and anchor_object_dbid to get the anchor object for the command.
        """
        # TODO: Is the anchor object type enough here, or do we need a mapping? The home-app model
        #  names might not exactly match the plugins model names.
        anchor_model = apps.get_model(
            app_label=self._meta.app_label, model_name=self.anchor_object_type
        )
        return anchor_model.objects.get(dbid=self.anchor_object_dbid)



from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from canvas_cli.apps.auth import get_or_request_api_token


@pytest.fixture
def valid_token_response() -> Any:
    """Returns a valid token response."""

    class TokenResponse:
        status_code = 200

        def json(self) -> dict:
            return {"access_token": "a-valid-api-token", "expires_in": 3600}

    return TokenResponse()


@pytest.fixture
def error_token_response() -> Any:
    """Returns an error token response."""

    class TokenResponse:
        status_code = 500

    return TokenResponse()


@pytest.fixture
def expired_token_response() -> Any:
    """Returns an expired token response."""

    class TokenResponse:
        status_code = 200

        def json(self) -> dict:
            return {"access_token": "a-valid-api-token", "expires_in": -1}

    return TokenResponse()


@patch("keyring.get_password")
@patch("requests.Session.post")
@patch("canvas_cli.apps.auth.utils.is_token_valid")
def test_get_or_request_api_token_uses_stored_token(
    mock_is_token_valid: MagicMock,
    mock_post: MagicMock,
    mock_get_password: MagicMock,
    valid_token_response: Any,
) -> None:
    """Test that get_or_request_api_token uses a stored token if it is valid."""
    mock_is_token_valid.return_value = True
    mock_get_password.return_value = "a-stored-valid-token"
    mock_post.return_value = valid_token_response

    token = get_or_request_api_token("http://localhost:8000")

    assert token == "a-stored-valid-token"
    mock_post.assert_not_called()


@patch("keyring.set_password")
@patch("keyring.get_password")
@patch("requests.Session.post")
@patch("canvas_cli.apps.auth.utils.get_api_client_credentials")
def test_get_or_request_api_token_requests_token_if_none_stored(
    mock_client_credentials: MagicMock,
    mock_post: MagicMock,
    mock_get_password: MagicMock,
    mock_set_password: MagicMock,
    valid_token_response: Any,
) -> None:
    """Test that get_or_request_api_token requests a new token if none is stored."""
    mock_client_credentials.return_value = "client_id=id&client_secret=secret"
    mock_get_password.return_value = None
    mock_post.return_value = valid_token_response

    token = get_or_request_api_token("http://localhost:8000")

    assert token == "a-valid-api-token"
    mock_post.assert_called_once_with(
        "http://localhost:8000/auth/token/",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        json=None,
        data="grant_type=client_credentials&scope=system/Plugins.*&client_id=id&client_secret=secret",
    )
    mock_set_password.assert_called_with(
        "canvas_cli.apps.auth.utils",
        username="http://localhost:8000|token",
        password="a-valid-api-token",
    )


@patch("keyring.get_password")
@patch("requests.Session.post")
@patch("canvas_cli.apps.auth.utils.get_api_client_credentials")
def test_get_or_request_api_token_raises_exception_if_error_token_response(
    mock_client_credentials: MagicMock,
    mock_post: MagicMock,
    mock_get_password: MagicMock,
    error_token_response: Any,
) -> None:
    """Test that get_or_request_api_token raises an exception if an error token response is received."""
    mock_client_credentials.return_value = "client_id=id&client_secret=secret"
    mock_get_password.return_value = None
    mock_post.return_value = error_token_response

    with pytest.raises(Exception) as e:
        get_or_request_api_token("http://localhost:8000")

    assert "Unable to get a valid access token from the given host 'http://localhost:8000'" in repr(
        e
    )

    mock_post.assert_called_once_with(
        "http://localhost:8000/auth/token/",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        json=None,
        data="grant_type=client_credentials&scope=system/Plugins.*&client_id=id&client_secret=secret",
    )


@patch("keyring.get_password")
@patch("requests.Session.post")
@patch("canvas_cli.apps.auth.utils.get_api_client_credentials")
def test_get_or_request_api_token_raises_exception_if_expired_token(
    mock_client_credentials: MagicMock,
    mock_post: MagicMock,
    mock_get_password: MagicMock,
    expired_token_response: Any,
) -> None:
    """Test that get_or_request_api_token raises an exception if an expired token is received."""
    mock_client_credentials.return_value = "client_id=id&client_secret=secret"
    mock_get_password.return_value = None
    mock_post.return_value = expired_token_response

    with pytest.raises(Exception) as e:
        get_or_request_api_token("http://localhost:8000")

    assert (
        "A valid token could not be acquired from the given host 'http://localhost:8000'" in repr(e)
    )

    mock_post.assert_called_once_with(
        "http://localhost:8000/auth/token/",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        json=None,
        data="grant_type=client_credentials&scope=system/Plugins.*&client_id=id&client_secret=secret",
    )



# Normally we would register Django models in the models module, but they are in "data" for legacy
# reasons. For this reason, we import * from data to avoid duplication.

from .data import *  # noqa: F403



from enum import StrEnum
from typing import Any, Self

from pydantic import model_validator

from canvas_sdk.effects import EffectType, _BaseEffect


class LaunchModalEffect(_BaseEffect):
    """An Effect that will launch a modal."""

    class Meta:
        effect_type = EffectType.LAUNCH_MODAL

    class TargetType(StrEnum):
        DEFAULT_MODAL = "default_modal"
        NEW_WINDOW = "new_window"
        RIGHT_CHART_PANE = "right_chart_pane"

    url: str | None = None
    content: str | None = None
    target: TargetType = TargetType.DEFAULT_MODAL

    @property
    def values(self) -> dict[str, Any]:
        """The LaunchModalEffect values."""
        return {"url": self.url, "content": self.content, "target": self.target.value}

    @model_validator(mode="after")
    def check_mutually_exclusive_fields(self) -> Self:
        """Check that url and content are mutually exclusive."""
        if self.url is not None and self.content is not None:
            raise ValueError("'url' and 'content' are mutually exclusive")

        return self



from typing import Any

from pydantic import Field

from canvas_generated.messages.effects_pb2 import EffectType
from canvas_sdk.effects.base import _BaseEffect


class ShowButtonEffect(_BaseEffect):
    """
    An Effect that will decide an action button's properties.
    """

    class Meta:
        effect_type = EffectType.SHOW_ACTION_BUTTON

    key: str = Field(min_length=1)
    title: str = Field(min_length=1)

    @property
    def values(self) -> dict[str, Any]:
        """The ShowButtonEffect's values."""
        return {"key": self.key, "title": self.title}

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}



from .allergy_intolerance import AllergyIntolerance, AllergyIntoleranceCoding
from .appointment import Appointment
from .assessment import Assessment
from .billing import BillingLineItem
from .care_team import CareTeamMembership, CareTeamRole
from .command import Command
from .condition import Condition, ConditionCoding
from .coverage import Coverage, Transactor, TransactorAddress, TransactorPhone
from .detected_issue import DetectedIssue, DetectedIssueEvidence
from .device import Device
from .imaging import ImagingOrder, ImagingReport, ImagingReview
from .lab import (
    LabOrder,
    LabOrderReason,
    LabOrderReasonCondition,
    LabReport,
    LabReview,
    LabTest,
    LabValue,
    LabValueCoding,
)
from .medication import Medication, MedicationCoding
from .note import Note, NoteType
from .observation import (
    Observation,
    ObservationCoding,
    ObservationComponent,
    ObservationComponentCoding,
    ObservationValueCoding,
)
from .organization import Organization
from .patient import (
    Patient,
    PatientAddress,
    PatientContactPoint,
    PatientExternalIdentifier,
    PatientSetting,
)
from .practicelocation import PracticeLocation, PracticeLocationSetting
from .protocol_override import ProtocolOverride
from .questionnaire import (
    Interview,
    InterviewQuestionnaireMap,
    InterviewQuestionResponse,
    Question,
    Questionnaire,
    QuestionnaireQuestionMap,
    ResponseOption,
    ResponseOptionSet,
)
from .staff import Staff
from .task import Task, TaskComment, TaskLabel, TaskTaskLabel
from .user import CanvasUser

__all__ = [
    "Appointment",
    "AllergyIntolerance",
    "AllergyIntoleranceCoding",
    "Assessment",
    "BillingLineItem",
    "CanvasUser",
    "CareTeamMembership",
    "CareTeamRole",
    "Command",
    "Condition",
    "ConditionCoding",
    "Coverage",
    "DetectedIssue",
    "DetectedIssueEvidence",
    "Device",
    "ImagingOrder",
    "ImagingReport",
    "ImagingReview",
    "Interview",
    "InterviewQuestionnaireMap",
    "InterviewQuestionResponse",
    "LabOrder",
    "LabOrderReason",
    "LabOrderReasonCondition",
    "LabReport",
    "LabReview",
    "LabTest",
    "LabValue",
    "LabValueCoding",
    "Medication",
    "MedicationCoding",
    "Note",
    "NoteType",
    "Observation",
    "ObservationCoding",
    "ObservationComponent",
    "ObservationComponentCoding",
    "ObservationValueCoding",
    "Organization",
    "Patient",
    "PatientAddress",
    "PatientContactPoint",
    "PatientExternalIdentifier",
    "PatientSetting",
    "PracticeLocation",
    "PracticeLocationSetting",
    "ProtocolOverride",
    "Question",
    "Questionnaire",
    "QuestionnaireQuestionMap",
    "ResponseOption",
    "ResponseOptionSet",
    "Staff",
    "Task",
    "TaskComment",
    "TaskLabel",
    "TaskTaskLabel",
    "Transactor",
    "TransactorAddress",
    "TransactorPhone",
]



import pytest

from canvas_cli.utils.urls import CoreEndpoint


@pytest.mark.parametrize("path", ["a-path", "/a-path", "/a-path/", "a-path/"])
def test_endpoint_builder(path: str) -> None:
    """Test that the endpoint is always generated with a trailing `/`."""
    assert (
        CoreEndpoint.LOG.build("https://test-host.com", path)
        == "https://test-host.com/core/api/v1/logging/a-path/"
    )



from canvas_sdk.value_set.v2022.condition import (
    DisordersOfTheImmuneSystem,
    EncephalopathyDueToChildhoodVaccination,
    Rhabdomyolysis,
    StableAndUnstableAngina,
)
from canvas_sdk.value_set.value_set import CombinedValueSet


def test_value_set_class_values_property() -> None:
    """Test that the value_set returns the correct values."""
    value_set = DisordersOfTheImmuneSystem
    assert value_set.values["ICD10CM"] == DisordersOfTheImmuneSystem.ICD10CM
    assert value_set.values["SNOMEDCT"] == DisordersOfTheImmuneSystem.SNOMEDCT


def test_value_set_class_pipe_operator_with_two_value_sets() -> None:
    """Test that the pipe operator returns the correct values."""
    combined_value_set: CombinedValueSet = (
        DisordersOfTheImmuneSystem | EncephalopathyDueToChildhoodVaccination
    )

    both_classes_icd_10_codes = DisordersOfTheImmuneSystem.ICD10CM.union(
        EncephalopathyDueToChildhoodVaccination.ICD10CM
    )
    both_classes_snomed_codes = DisordersOfTheImmuneSystem.SNOMEDCT.union(
        EncephalopathyDueToChildhoodVaccination.SNOMEDCT
    )

    assert both_classes_icd_10_codes == combined_value_set.values["ICD10CM"]
    assert both_classes_snomed_codes == combined_value_set.values["SNOMEDCT"]


def test_value_set_class_pipe_operator_with_three_value_sets() -> None:
    """Test that the pipe operator returns the correct values with multiple operands."""
    combined_value_set: CombinedValueSet = (
        DisordersOfTheImmuneSystem | EncephalopathyDueToChildhoodVaccination | Rhabdomyolysis
    )

    all_classes_icd_10_codes = DisordersOfTheImmuneSystem.ICD10CM.union(
        EncephalopathyDueToChildhoodVaccination.ICD10CM
    ).union(Rhabdomyolysis.ICD10CM)
    all_classes_snomed_codes = DisordersOfTheImmuneSystem.SNOMEDCT.union(
        EncephalopathyDueToChildhoodVaccination.SNOMEDCT
    ).union(Rhabdomyolysis.SNOMEDCT)

    assert all_classes_icd_10_codes == combined_value_set.values["ICD10CM"]
    assert all_classes_snomed_codes == combined_value_set.values["SNOMEDCT"]


def test_value_set_class_pipe_operator_with_two_combined_value_sets() -> None:
    """Test that the pipe operator returns the correct values with two combined value_sets."""
    combined_value_set_1: CombinedValueSet = (
        DisordersOfTheImmuneSystem | EncephalopathyDueToChildhoodVaccination
    )
    combined_value_set_2: CombinedValueSet = Rhabdomyolysis | StableAndUnstableAngina

    combined_value_set = combined_value_set_1 | combined_value_set_2

    all_classes_icd_10_codes = (
        DisordersOfTheImmuneSystem.ICD10CM.union(EncephalopathyDueToChildhoodVaccination.ICD10CM)
        .union(Rhabdomyolysis.ICD10CM)
        .union(StableAndUnstableAngina.ICD10CM)
    )
    all_classes_snomed_codes = (
        DisordersOfTheImmuneSystem.SNOMEDCT.union(EncephalopathyDueToChildhoodVaccination.SNOMEDCT)
        .union(Rhabdomyolysis.SNOMEDCT)
        .union(StableAndUnstableAngina.SNOMEDCT)
    )

    assert combined_value_set.values["ICD10CM"] == all_classes_icd_10_codes
    assert combined_value_set.values["SNOMEDCT"] == all_classes_snomed_codes



import time
from collections.abc import Callable, Mapping
from functools import wraps
from typing import Any, TypeVar, cast

import requests
import statsd

F = TypeVar("F", bound=Callable)


class Http:
    """A helper class for completing HTTP calls with metrics tracking."""

    def __init__(self) -> None:
        self.session = requests.Session()
        self.statsd_client = statsd.StatsClient()

    @staticmethod
    def measure_time(fn: F) -> F:
        """A decorator to store timing of HTTP calls."""

        @wraps(fn)
        def wrapper(self: "Http", *args: Any, **kwargs: Any) -> Any:
            start_time = time.time()
            result = fn(self, *args, **kwargs)
            end_time = time.time()
            timing = int((end_time - start_time) * 1000)
            self.statsd_client.timing(f"http_{fn.__name__}", timing)
            return result

        return cast(F, wrapper)

    @measure_time
    def get(
        self, url: str, headers: Mapping[str, str | bytes | None] | None = None
    ) -> requests.Response:
        """Sends a GET request."""
        if headers is None:
            headers = {}
        return self.session.get(url, headers=headers)

    @measure_time
    def post(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a POST request."""
        return self.session.post(url, json=json, data=data, headers=headers)

    @measure_time
    def put(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a PUT request."""
        return self.session.put(url, json=json, data=data, headers=headers)

    @measure_time
    def patch(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a PATCH request."""
        return self.session.patch(url, json=json, data=data, headers=headers)



from unittest.mock import MagicMock, patch

from canvas_sdk.utils import Http


@patch("requests.Session.get")
def test_http_get(mock_get: MagicMock) -> None:
    """Test that the Http.get method calls requests.get with the correct arguments."""
    http = Http()
    http.get("https://www.canvasmedical.com/", headers={"Authorization": "Bearer as;ldkfjdkj"})
    mock_get.assert_called_once_with(
        "https://www.canvasmedical.com/", headers={"Authorization": "Bearer as;ldkfjdkj"}
    )


@patch("requests.Session.post")
def test_http_post(mock_post: MagicMock) -> None:
    """Test that the Http.post method calls requests.post with the correct arguments."""
    http = Http()
    http.post(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )
    mock_post.assert_called_once_with(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )


@patch("requests.Session.put")
def test_http_put(mock_put: MagicMock) -> None:
    """Test that the Http.put method calls requests.put with the correct arguments."""
    http = Http()
    http.put(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )
    mock_put.assert_called_once_with(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )


@patch("requests.Session.patch")
def test_http_patch(mock_patch: MagicMock) -> None:
    """Test that the Http.patch method calls requests.patch with the correct arguments."""
    http = Http()
    http.patch(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )
    mock_patch.assert_called_once_with(
        "https://www.canvasmedical.com/",
        json={"hey": "hi"},
        data="grant-type=client_credentials",
        headers={"Content-type": "application/json"},
    )



from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class UpdateDiagnosisCommand(BaseCommand):
    """A class for managing an Update Diagnosis command within a specific note."""

    class Meta:
        key = "updateDiagnosis"
        commit_required_fields = (
            "condition_code",
            "new_condition_code",
        )

    condition_code: str | None = None
    new_condition_code: str | None = None
    background: str | None = None
    narrative: str | None = None

    @property
    def values(self) -> dict:
        """The Update Diagnosis command's field values."""
        return {
            "condition_code": self.condition_code,
            "new_condition_code": self.new_condition_code,
            "background": self.background,
            "narrative": self.narrative,
        }



import inspect
from pathlib import Path
from typing import Any

from django.template import Context, Template

from settings import PLUGIN_DIRECTORY


def render_to_string(template_name: str, context: dict[str, Any] | None = None) -> str | None:
    """Load a template and render it with the given context.

    Args:
        template_name (str): The path to the template file, relative to the plugin package.
            If the path starts with a forward slash ("/"), it will be stripped during resolution.
        context (dict[str, Any] | None): A dictionary of variables to pass to the template
            for rendering. Defaults to None, which uses an empty context.

    Returns:
        str: The rendered template as a string.

    Raises:
        FileNotFoundError: If the template file does not exist within the plugin's directory
            or if the resolved path is invalid.
    """
    plugins_dir = Path(PLUGIN_DIRECTORY).resolve()
    current_frame = inspect.currentframe()
    caller = current_frame.f_back if current_frame else None

    if not caller or "__is_plugin__" not in caller.f_globals:
        return None

    plugin_name = caller.f_globals["__name__"].split(".")[0]
    plugin_dir = plugins_dir / plugin_name
    template_path = Path(plugin_dir / template_name.lstrip("/")).resolve()

    if not template_path.is_relative_to(plugin_dir):
        raise PermissionError(f"Invalid template '{template_name}'")
    elif not template_path.exists():
        raise FileNotFoundError(f"Template {template_name} not found.")

    template = Template(template_path.read_text())

    return template.render(Context(context))



from canvas_sdk.commands.commands.questionnaire import QuestionnaireCommand


class PhysicalExamCommand(QuestionnaireCommand):
    """A class for managing physical exam command."""

    class Meta:
        key = "exam"
        commit_required_fields = ("questionnaire_id",)



from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class QuestionnaireCommand(_BaseCommand):
    """A class for managing a Questionnaire command within a specific note."""

    class Meta:
        key = "questionnaire"
        commit_required_fields = ("questionnaire_id",)

    questionnaire_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "questionnaire"}
    )
    result: str | None = None

    @property
    def values(self) -> dict:
        """The Questionnaire command's field values."""
        return {"questionnaire_id": self.questionnaire_id, "result": self.result}



from datetime import date

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class DiagnoseCommand(_BaseCommand):
    """A class for managing a Diagnose command within a specific note."""

    class Meta:
        key = "diagnose"
        commit_required_fields = ("icd10_code",)

    icd10_code: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "diagnose"}
    )
    background: str | None = None
    approximate_date_of_onset: date | None = None
    today_assessment: str | None = None

    @property
    def values(self) -> dict:
        """The Diagnose command's field values."""
        return {
            "icd10_code": self.icd10_code,
            "background": self.background,
            "approximate_date_of_onset": (
                self.approximate_date_of_onset.isoformat()
                if self.approximate_date_of_onset
                else None
            ),
            "today_assessment": self.today_assessment,
        }



from typing import NotRequired

from typing_extensions import TypedDict


class Coding(TypedDict):
    """Coding object in Canvas."""

    system: str
    code: str
    display: NotRequired[str]


class ClinicalQuantity(TypedDict):
    """A ClinicalQuantity for prescribe/refill commands."""

    representative_ndc: str
    ncpdp_quantity_qualifier_code: str



import logging
import os
from typing import Any

from pubsub.pubsub import Publisher


class PubSubLogHandler(logging.Handler):
    """Custom logging handler that publishes logs to a pub/sub channel."""

    def __init__(self) -> None:
        self.publisher = Publisher()
        logging.Handler.__init__(self=self)

    def emit(self, record: Any) -> None:
        """Publishes the log message to the pub/sub channel."""
        message = self.format(record)
        self.publisher.publish(message)


class PluginLogger:
    """A custom logger for plugins."""

    def __init__(self) -> None:
        self.logger = logging.getLogger("plugin_runner_logger")
        self.logger.setLevel(logging.INFO)
        log_prefix = os.getenv("HOSTNAME", "")
        if log_prefix != "":
            log_prefix = f"[{log_prefix}] "
        formatter = logging.Formatter(f"{log_prefix}%(levelname)s %(asctime)s %(message)s")

        streaming_handler = logging.StreamHandler()
        streaming_handler.setFormatter(formatter)
        self.logger.addHandler(streaming_handler)

        if os.getenv("REDIS_ENDPOINT"):
            pubsub_handler = PubSubLogHandler()
            pubsub_handler.setFormatter(formatter)
            self.logger.addHandler(pubsub_handler)

    def debug(self, message: Any) -> None:
        """Logs a debug message."""
        self.logger.debug(message)

    def info(self, message: Any) -> None:
        """Logs an info message."""
        self.logger.info(message)

    def warning(self, message: Any) -> None:
        """Logs a warning message."""
        self.logger.warning(message)

    def error(self, message: Any) -> None:
        """Logs an error message."""
        self.logger.error(message)

    def critical(self, message: Any) -> None:
        """Logs a critical message."""
        self.logger.critical(message)



import pytest

from plugin_runner.sandbox import FORBIDDEN_ASSIGNMENTS, Sandbox

# Sample code strings for testing various scenarios
VALID_CODE = """
x = 10
y = 20
result = x + y
"""

CODE_WITH_RESTRICTED_IMPORT = """
import os
result = os.listdir('.')
"""

CODE_WITH_PLUGIN_RUNNER_SETTING_IMPORT = """
import settings
result = settings.AWS_SECRET_ACCESS_KEY
"""

CODE_WITH_ALLOWED_IMPORT = """
import json
result = json.dumps({"key": "value"})
"""

CODE_WITH_FORBIDDEN_FUNC_NAME = """
builtins = {}
"""

SOURCE_CODE_MODULE = """
import module.b
result = module.b
"""

CODE_WITH_FORBIDDEN_ASSIGNMENTS = [
    code
    for var in FORBIDDEN_ASSIGNMENTS
    for code in [
        f"{var} = 'test'",
        f"test = {var} = 'test'",
        f"test = {var} = test2 = 'test'",
        f"(a, (b, c), (d, ({var}, f))) = (1, (2, 3), (4, (5, 6)))",
        f"(a, (b, c), (d, [{var}, f])) = (1, (2, 3), (4, [5, 6]))",
    ]
]


def test_valid_code_execution() -> None:
    """Test execution of valid code in the sandbox."""
    sandbox = Sandbox(VALID_CODE)
    scope = sandbox.execute()
    assert scope["result"] == 30, "The code should compute result as 30."


def test_disallowed_import() -> None:
    """Test that restricted imports are not allowed."""
    sandbox = Sandbox(CODE_WITH_RESTRICTED_IMPORT)
    with pytest.raises(ImportError, match="'os' is not an allowed import."):
        sandbox.execute()


def test_plugin_runner_settings_import() -> None:
    """Test that imports of plugin runner settings are not allowed."""
    sandbox = Sandbox(CODE_WITH_PLUGIN_RUNNER_SETTING_IMPORT)
    with pytest.raises(ImportError, match="'settings' is not an allowed import."):
        sandbox.execute()


def test_allowed_import() -> None:
    """Test that allowed imports (from ALLOWED_MODULES) work correctly."""
    sandbox = Sandbox(CODE_WITH_ALLOWED_IMPORT)
    scope = sandbox.execute()
    assert scope["result"] == '{"key": "value"}', "JSON encoding should work with allowed imports."


def test_forbidden_name() -> None:
    """Test that forbidden function names are blocked by Transformer."""
    sandbox = Sandbox(CODE_WITH_FORBIDDEN_FUNC_NAME)
    with pytest.raises(RuntimeError, match="Code is invalid"):
        sandbox.execute()


@pytest.mark.parametrize("code", CODE_WITH_FORBIDDEN_ASSIGNMENTS)
def test_forbidden_assignment(code: str) -> None:
    """Test that forbidden assignments are blocked by Transformer."""
    sandbox = Sandbox(code)
    with pytest.raises(RuntimeError, match="Code is invalid"):
        sandbox.execute()


def test_code_with_warnings() -> None:
    """Test that the sandbox captures warnings for restricted names or usage."""
    code_with_warning = """
_x = 5
result = _x
"""
    sandbox = Sandbox(code_with_warning)
    assert sandbox.warnings, "There should be warnings for using restricted names."
    scope = sandbox.execute()
    assert scope["result"] == 5, "Code should execute despite warnings."


def test_compile_errors() -> None:
    """Test that compile errors are detected for invalid syntax."""
    invalid_code = """
def missing_colon()
    return 42
"""
    sandbox = Sandbox(invalid_code)
    with pytest.raises(RuntimeError, match="Code is invalid"):
        sandbox.execute()


def test_sandbox_scope() -> None:
    """Verify the sandbox scope includes expected built-ins and utility functions."""
    sandbox = Sandbox(VALID_CODE)
    scope = sandbox.execute()
    assert "any" in scope["__builtins__"]
    assert scope["__builtins__"]["any"] == any, "'any' function should be accessible in sandbox."


def test_print_collector() -> None:
    """Ensure that PrintCollector is used for capturing prints."""
    code_with_print = """
print("Hello, Sandbox!")
"""
    sandbox = Sandbox(code_with_print)
    scope = sandbox.execute()
    assert "Hello, Sandbox!" in scope["_print"].txt, "Print output should be captured."


def test_sandbox_denies_module_name_import_outside_package() -> None:
    """Test that modules outside the root package cannot be imported."""
    sandbox_module_a = Sandbox(source_code=SOURCE_CODE_MODULE, namespace="other_module.a")
    with pytest.raises(ImportError, match="module.b' is not an allowed import."):
        sandbox_module_a.execute()



from abc import abstractmethod
from collections.abc import Container
from typing import TYPE_CHECKING, Any, Protocol, Self, cast

from django.db import models
from django.db.models import Q

if TYPE_CHECKING:
    from canvas_sdk.protocols.timeframe import Timeframe
    from canvas_sdk.value_set.value_set import ValueSet


class BaseModelManager(models.Manager):
    """A base manager for models."""

    def get_queryset(self) -> models.QuerySet:
        """Return a queryset that filters out deleted objects."""
        return super().get_queryset().filter(deleted=False)


class BaseQuerySet(models.QuerySet):
    """A base QuerySet inherited from Django's model.Queryset."""

    pass


class QuerySetProtocol(Protocol):
    """A typing protocol for use in mixins into models.QuerySet-inherited classes."""

    def filter(self, *args: Any, **kwargs: Any) -> Self:
        """Django's models.QuerySet filter method."""
        ...

    def distinct(self) -> Self:
        """Django's models.QuerySet distinct method."""
        ...


class ValueSetLookupQuerySetProtocol(QuerySetProtocol):
    """A typing protocol for use in mixins using value set lookup methods."""

    @staticmethod
    @abstractmethod
    def codings(value_set: type["ValueSet"]) -> tuple[tuple[str, set[str]]]:
        """A protocol method for defining codings."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def q_object(system: str, codes: Container[str]) -> Q:
        """A protocol method for defining Q objects for value set lookups."""
        raise NotImplementedError


class CommittableQuerySetMixin(QuerySetProtocol):
    """A queryset for committable objects."""

    def committed(self) -> Self:
        """Return a queryset that filters for objects that have been committed."""
        return self.filter(committer_id__isnull=False, entered_in_error_id__isnull=True)


class ForPatientQuerySetMixin(QuerySetProtocol):
    """A queryset for patient assets."""

    def for_patient(self, patient_id: str) -> Self:
        """Return a queryset that filters objects for a specific patient."""
        return self.filter(patient__id=patient_id)


class ValueSetLookupQuerySetMixin(ValueSetLookupQuerySetProtocol):
    """A QuerySet mixin that can filter objects based on a ValueSet."""

    def find(self, value_set: type["ValueSet"]) -> Self:
        """
        Filters conditions, medications, etc. to those found in the inherited ValueSet class that is passed.

        For example:

        from canvas_sdk.v1.data.condition import Condition
        from canvas_sdk.value_set.v2022.condition import MorbidObesity
        morbid_obesity_conditions = Condition.objects.find(MorbidObesity)

        This method can also be chained like so:

        Condition.objects.find(MorbidObesity).find(AnaphylacticReactionToCommonBakersYeast)
        """
        q_filter = Q()
        for system, codes in self.codings(value_set):
            q_filter |= self.q_object(system, codes)
        return self.filter(q_filter).distinct()

    @staticmethod
    def codings(value_set: type["ValueSet"]) -> tuple[tuple[str, set[str]]]:
        """Provide a sequence of tuples where each tuple is a code system URL and a set of codes."""
        values_dict = cast(dict, value_set.values)
        return cast(
            tuple[tuple[str, set[str]]],
            tuple(
                (i[1], values_dict[i[0]])
                for i in value_set.CODE_SYSTEM_MAPPING.items()
                if i[0] in values_dict
            ),
        )

    @staticmethod
    def q_object(system: str, codes: Container[str]) -> Q:
        """
        This method can be overridden if a Q object with different filtering options is needed.
        """
        return Q(codings__system=system, codings__code__in=codes)


class ValueSetLookupByNameQuerySetMixin(ValueSetLookupQuerySetMixin):
    """
    QuerySet for ValueSet lookups using code system name rather than URL.

    Some models, like Questionnaire, store the code system by name (e.g. "LOINC") rather than by the
    url (e.g. "http://loinc.org"). This subclass accommodates these models.
    """

    @staticmethod
    def codings(value_set: type["ValueSet"]) -> tuple[tuple[str, set[str]]]:
        """
        Provide a sequence of tuples where each tuple is a code system name and a set of codes.
        """
        values_dict = cast(dict, value_set.values)
        return cast(
            tuple[tuple[str, set[str]]],
            tuple(
                (i[0], values_dict[i[0]])
                for i in value_set.CODE_SYSTEM_MAPPING.items()
                if i[0] in values_dict
            ),
        )


class TimeframeLookupQuerySetProtocol(QuerySetProtocol):
    """A typing protocol for use in TimeframeLookupQuerySetMixin."""

    @property
    @abstractmethod
    def timeframe_filter_field(self) -> str:
        """A protocol method for timeframe_filter_field."""
        raise NotImplementedError


class TimeframeLookupQuerySetMixin(TimeframeLookupQuerySetProtocol):
    """A class that adds queryset functionality to filter using timeframes."""

    @property
    def timeframe_filter_field(self) -> str:
        """Returns the field that should be filtered on. Can be overridden for different models."""
        return "note__datetime_of_service"

    def within(self, timeframe: "Timeframe") -> Self:
        """A method to filter a queryset for datetimes within a timeframe."""
        return self.filter(
            **{
                f"{self.timeframe_filter_field}__range": (
                    timeframe.start.datetime,
                    timeframe.end.datetime,
                )
            }
        )


class CommittableQuerySet(BaseQuerySet, CommittableQuerySetMixin):
    """A queryset for committable objects."""

    pass


class ValueSetLookupQuerySet(BaseQuerySet, ValueSetLookupQuerySetMixin):
    """A class that includes methods for looking up value sets."""

    pass


class ValueSetLookupByNameQuerySet(BaseQuerySet, ValueSetLookupByNameQuerySetMixin):
    """A class that includes methods for looking up value sets by name."""

    pass


class ValueSetTimeframeLookupQuerySet(ValueSetLookupQuerySet, TimeframeLookupQuerySetMixin):
    """A class that includes methods for looking up value sets and using timeframes."""

    pass



from django.contrib.postgres.fields import ArrayField
from django.db import models

from canvas_sdk.v1.data.common import (
    AddressState,
    AddressType,
    AddressUse,
    ContactPointState,
    ContactPointSystem,
    ContactPointUse,
)


class CoverageStack(models.TextChoices):
    """CoverageStack."""

    IN_USE = "IN_USE", "In use"
    OTHER = "OTHER", "Other"
    REMOVED = "REMOVED", "Removed"


class CoverageState(models.TextChoices):
    """CoverageState."""

    ACTIVE = "active", "Active"
    DELETED = "deleted", "Deleted"


class CoverageType(models.TextChoices):
    """CoverageType."""

    COMMERCIAL = "commercial", "Commercial"
    WORKERS_COMP = "workerscomp", "Workers Comp"
    BCBS = "bcbs", "Blue Cross Blue Shield"
    TRICARE = "champus", "Tricare/Champus"
    MEDICAID = "medicaid", "Medicaid"
    MEDICARE = "medicare", "Medicare"
    OTHER = "other", "Other"
    TPA = "tpa", "Third Party Administrator"
    MVA = "motorvehicle", "Motor Vehicle"
    LIEN = "lien", "Attorney/Lien"
    PIP = "pip", "Personal Injury"


class CoverageRelationshipCode(models.TextChoices):
    """CoverageRelationshipCode."""

    SELF = "18", "Self"
    SPOUSE = "01" "Spouse"
    CHILD_INSURED_HAS_FINANCIAL_RESP = "19", "Natural Child, insured has financial responsibility"
    CHILD_HAS_FINANCIAL_RESP = "43", "Natural Child, insured does not have financial responsibility"
    STEP_CHILD = "17", "Step Child"
    FOSTER_CHILD = "10", "Foster Child"
    WARD_OF_COURT = "15", "Ward of the Court"
    EMPLOYEE = "20", "Employee"
    UNKNOWN = "21", "Unknown"
    HANDICAPPED_DEPENDENT = "22", "Handicapped Dependent"
    ORGAN_DONOR = "39", "Organ donor"
    CADAVER_DONOR = "40", "Cadaver donor"
    GRANDCHILD = "05", "Grandchild"
    NIECE_NEPHEW = "07", "Niece/Nephew"
    INJURED_PLAINTIFF = "41", "Injured Plaintiff"
    SPONSORED_DEPENDENT = "23", "Sponsored Dependent"
    MINOR_DEP_OF_MINOR_DEP = "24", "Minor Dependent of a Minor Dependent"
    MOTHER = "32", "Mother"
    FATHER = "33", "Father"
    GRANDPARENT = "04", "Grandparent"
    LIFE_PARTNER = "53", "Life Partner"
    SIGNIFICANT_OTHER = "29", "Significant Other"
    OTHER = "G8", "Other"


class TransactorCoverageType(models.TextChoices):
    """TransactorCoverageType."""

    ANNU = "ANNU", "annuity policy"
    AUTOPOL = "AUTOPOL", "automobile"
    CHAR = "CHAR", "charity program"
    COL = "COL", "collision coverage policy"
    CRIME = "CRIME", "crime victim program"
    DENTAL = "DENTAL", "dental care policy"
    DENTPRG = "DENTPRG", "dental program"
    DIS = "DIS", "disability insurance policy"
    DISEASE = "DISEASE", "disease specific policy"
    DRUGPOL = "DRUGPOL", "drug policy"
    EAP = "EAP", "employee assistance program"
    EWB = "EWB", "employee welfare benefit plan policy"
    ENDRENAL = "ENDRENAL", "end renal program"
    EHCPOL = "EHCPOL", "extended healthcare"
    FLEXP = "FLEXP", "flexible benefit plan policy"
    GOVEMP = "GOVEMP", "government employee health program"
    HIP = "HIP", "health insurance plan policy"
    HMO = "HMO", "health maintenance organization policy"
    HSAPOL = "HSAPOL", "health spending account"
    HIRISK = "HIRISK", "high risk pool program"
    HIVAIDS = "HIVAIDS", "HIV-AIDS program"
    IND = "IND", "indigenous peoples health program"
    LIFE = "LIFE", "life insurance policy"
    LTC = "LTC", "long term care policy"
    MCPOL = "MCPOL", "managed care policy"
    MANDPOL = "MANDPOL", "mandatory health program"
    MENTPOL = "MENTPOL", "mental health policy"
    MENTPRG = "MENTPRG", "mental health program"
    MILITARY = "MILITARY", "military health program"
    PAY = "pay", "Pay"
    POS = "POS", "point of service policy"
    PPO = "PPO", "preferred provider organization policy"
    PNC = "PNC", "property and casualty insurance policy"
    DISEASEPRG = "DISEASEPRG", "public health program"
    PUBLICPOL = "PUBLICPOL", "public healthcare"
    REI = "REI", "reinsurance policy"
    RETIRE = "RETIRE", "retiree health program"
    SAFNET = "SAFNET", "safety net clinic program"
    SOCIAL = "SOCIAL", "social service program"
    SUBSIDIZ = "SUBSIDIZ", "subsidized health program"
    SUBSIDMC = "SUBSIDMC", "subsidized managed care program"
    SUBSUPP = "SUBSUPP", "subsidized supplemental health program"
    SUBPOL = "SUBPOL", "substance use policy"
    SUBPRG = "SUBPRG", "substance use program"
    SURPL = "SURPL", "surplus line insurance policy"
    TLIFE = "TLIFE", "term life insurance policy"
    UMBRL = "UMBRL", "umbrella liability insurance policy"
    UNINSMOT = "UNINSMOT", "uninsured motorist policy"
    ULIFE = "ULIFE", "universal life insurance policy"
    VET = "VET", "veteran health program"
    VISPOL = "VISPOL", "vision care policy"
    CANPRG = "CANPRG", "women's cancer detection program"
    WCBPOL = "WCBPOL", "worker's compensation"


class TransactorType(models.TextChoices):
    """TransactorType."""

    COMMERCIAL = "commercial", "Commercial"
    WORKERS_COMP = "workerscomp", "Workers Comp"
    TRICARE = "champus", "Tricare/Champus"
    MEDICAID = "medicaid", "Medicaid"
    MEDICARE = "medicare", "Medicare"
    MEDICARE_ADVANTAGE = "medicare_advantage", "Medicare Advantage"
    CHIP = "CHIP", "CHIP"
    AUTO = "automobile", "Automobile"
    EMPLOYER = "employer", "Employer"
    DIRECT_CARE = "direct_care", "Direct Care"
    BCBS = "bcbs", "Blue Cross Blue Shield"


class Coverage(models.Model):
    """Coverage."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_coverage_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="coverages", null=True
    )
    guarantor = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="dependent_coverages", null=True
    )
    subscriber = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="subscribed_coverages", null=True
    )
    patient_relationship_to_subscriber = models.CharField(choices=CoverageRelationshipCode.choices)
    issuer = models.ForeignKey(
        "v1.Transactor", on_delete=models.DO_NOTHING, related_name="coverages", null=True
    )
    id_number = models.CharField()
    plan = models.CharField()
    sub_plan = models.CharField()
    group = models.CharField()
    sub_group = models.CharField()
    employer = models.CharField()
    coverage_start_date = models.DateField()
    coverage_end_date = models.DateField()
    coverage_rank = models.IntegerField()
    state = models.CharField(choices=CoverageState.choices)
    plan_type = models.CharField(choices=CoverageType.choices)
    coverage_type = models.CharField(choices=TransactorCoverageType.choices)
    issuer_address = models.ForeignKey(
        "v1.TransactorAddress",
        on_delete=models.DO_NOTHING,
        related_name="coverages",
        null=True,
    )
    issuer_phone = models.ForeignKey(
        "v1.TransactorPhone", on_delete=models.DO_NOTHING, related_name="coverages", null=True
    )
    comments = models.TextField()
    stack = models.CharField(choices=CoverageStack.choices)

    def __str__(self) -> str:
        return f"id={self.id}"


class Transactor(models.Model):
    """Transactor."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_quality_and_revenue_transactor_001"

    dbid = models.BigIntegerField(primary_key=True)
    payer_id = models.CharField()
    name = models.CharField()
    type = models.CharField()
    transactor_type = models.CharField(choices=TransactorType.choices)
    clearinghouse_payer = models.BooleanField()
    institutional = models.BooleanField(null=True)
    institutional_enrollment_req = models.BooleanField(null=True)
    professional = models.BooleanField(null=True)
    professional_enrollment_req = models.BooleanField(null=True)
    era = models.BooleanField(null=True)
    era_enrollment_req = models.BooleanField(null=True)
    eligibility = models.BooleanField(null=True)
    eligibility_enrollment_req = models.BooleanField(null=True)
    workers_comp = models.BooleanField(null=True)
    secondary_support = models.BooleanField(null=True)
    claim_fee = models.BooleanField(null=True)
    remit_fee = models.BooleanField(null=True)
    state = models.CharField()
    description = models.CharField()
    active = models.BooleanField()
    use_provider_for_eligibility = models.BooleanField()

    use_for_submission = models.ForeignKey(
        "v1.Transactor",
        on_delete=models.DO_NOTHING,
        null=True,
        related_name="used_for_submission_by",
    )

    coverage_types = ArrayField(models.CharField(choices=TransactorCoverageType.choices))


class TransactorAddress(models.Model):
    """TransactorAddress."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_quality_and_revenue_transactoraddress_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    line1 = models.CharField()
    line2 = models.CharField()
    city = models.CharField()
    district = models.CharField()
    state_code = models.CharField()
    postal_code = models.CharField()
    use = models.CharField(choices=AddressUse.choices)
    type = models.CharField(choices=AddressType.choices)
    longitude = models.FloatField(null=True, default=None, blank=True)
    latitude = models.FloatField(null=True, default=None, blank=True)
    start = models.DateField(null=True, blank=True)
    end = models.DateField(null=True, blank=True)
    country = models.CharField(max_length=255)
    state = models.CharField(choices=AddressState.choices)
    transactor = models.ForeignKey(
        "v1.Transactor", on_delete=models.DO_NOTHING, related_name="addresses", null=True
    )

    def __str__(self) -> str:
        return f"id={self.id}"


class TransactorPhone(models.Model):
    """TransactorPhone."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_quality_and_revenue_transactorphone_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    system = models.CharField(choices=ContactPointSystem.choices)
    value = models.CharField()
    use = models.CharField(choices=ContactPointUse.choices)
    use_notes = models.CharField()
    rank = models.IntegerField()
    state = models.CharField(choices=ContactPointState.choices)
    transactor = models.ForeignKey(
        "v1.Transactor", on_delete=models.DO_NOTHING, related_name="phones", null=True
    )

    def __str__(self) -> str:
        return f"id={self.id}"
